
class AzureInstancePowerState:
    RUNNING = 'PowerState/running'
    DEALLOCATED = 'PowerState/deallocated'
    STOPPED = 'PowerState/stopped'
    STARTING = 'PowerState/starting'
    DEALLOCATING = 'PowerState/deallocating'
    STOPPING = 'PowerState/stopping'
